import logging
import re
from typing import TYPE_CHECKING, Any, Tuple

import aiohttp_cors
import trafaret as t
from aiohttp import web

from ai.backend.common import msgpack
from ai.backend.logging import BraceStyleAdapter

from ..errors.exceptions import (
    DomainNotFound,
    DotfileAlreadyExists,
    DotfileCreationFailed,
    DotfileNotFound,
    GenericForbidden,
    InvalidAPIParameters,
)
from ..models import MAXIMUM_DOTFILE_SIZE, domains, query_domain_dotfiles, verify_dotfile_name
from .auth import admin_required, auth_required
from .manager import READ_ALLOWED, server_status_required
from .types import CORSOptions, Iterable, WebMiddleware
from .utils import check_api_params

if TYPE_CHECKING:
    from .context import RootContext

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


@server_status_required(READ_ALLOWED)
@admin_required
@check_api_params(
    t.Dict(
        {
            t.Key("domain"): t.String,
            t.Key("data"): t.String(max_length=MAXIMUM_DOTFILE_SIZE),
            t.Key("path"): t.String,
            t.Key("permission"): t.Regexp(r"^[0-7]{3}$", re.ASCII),
        },
    )
)
async def create(request: web.Request, params: Any) -> web.Response:
    log.info("DOMAINCOFNIG.CREATE_DOTFILE (domain: {0})", params["domain"])
    if not request["is_superadmin"] and request["user"]["domain_name"] != params["domain"]:
        raise GenericForbidden("Domain admins cannot create dotfiles of other domains")
    root_ctx: RootContext = request.app["_root.context"]
    async with root_ctx.db.begin() as conn:
        dotfiles, leftover_space = await query_domain_dotfiles(conn, params["domain"])
        if dotfiles is None:
            raise DomainNotFound("Input domain is not found")
        if leftover_space == 0:
            raise DotfileCreationFailed("No leftover space for dotfile storage")
        if len(dotfiles) == 100:
            raise DotfileCreationFailed("Dotfile creation limit reached")
        if not verify_dotfile_name(params["path"]):
            raise InvalidAPIParameters("dotfile path is reserved for internal operations.")

        duplicate = [x for x in dotfiles if x["path"] == params["path"]]
        if len(duplicate) > 0:
            raise DotfileAlreadyExists
        new_dotfiles = list(dotfiles)
        new_dotfiles.append({
            "path": params["path"],
            "perm": params["permission"],
            "data": params["data"],
        })
        dotfile_packed = msgpack.packb(new_dotfiles)
        if len(dotfile_packed) > MAXIMUM_DOTFILE_SIZE:
            raise DotfileCreationFailed("No leftover space for dotfile storage")

        query = (
            domains.update()
            .values(dotfiles=dotfile_packed)
            .where(domains.c.name == params["domain"])
        )
        await conn.execute(query)
    return web.json_response({})


@server_status_required(READ_ALLOWED)
@auth_required
@check_api_params(
    t.Dict({
        t.Key("domain"): t.String,
        t.Key("path", default=None): t.Null | t.String,
    })
)
async def list_or_get(request: web.Request, params: Any) -> web.Response:
    log.info("DOMAINCONFIG.LIST_OR_GET_DOTFILE (domain: {0})", params["domain"])
    if not request["is_superadmin"] and request["user"]["domain_name"] != params["domain"]:
        raise GenericForbidden("Users cannot access dotfiles of other domains")
    resp = []
    root_ctx: RootContext = request.app["_root.context"]
    async with root_ctx.db.begin() as conn:
        if params["path"]:
            dotfiles, _ = await query_domain_dotfiles(conn, params["domain"])
            if dotfiles is None:
                raise DomainNotFound
            for dotfile in dotfiles:
                if dotfile["path"] == params["path"]:
                    return web.json_response(dotfile)
            raise DotfileNotFound
        else:
            dotfiles, _ = await query_domain_dotfiles(conn, params["domain"])
            if dotfiles is None:
                raise DomainNotFound
            for entry in dotfiles:
                resp.append({
                    "path": entry["path"],
                    "permission": entry["perm"],
                    "data": entry["data"],
                })
            return web.json_response(resp)


@server_status_required(READ_ALLOWED)
@admin_required
@check_api_params(
    t.Dict(
        {
            t.Key("domain"): t.String,
            t.Key("data"): t.String(max_length=MAXIMUM_DOTFILE_SIZE),
            t.Key("path"): t.String,
            t.Key("permission"): t.Regexp(r"^[0-7]{3}$", re.ASCII),
        },
    )
)
async def update(request: web.Request, params: Any) -> web.Response:
    log.info("DOMAINCONFIG.UPDATE_DOTFILE (domain:{0})", params["domain"])
    if not request["is_superadmin"] and request["user"]["domain_name"] != params["domain"]:
        raise GenericForbidden("Domain admins cannot update dotfiles of other domains")
    root_ctx: RootContext = request.app["_root.context"]
    async with root_ctx.db.begin() as conn:
        dotfiles, _ = await query_domain_dotfiles(conn, params["domain"])
        if dotfiles is None:
            raise DomainNotFound
        new_dotfiles = [x for x in dotfiles if x["path"] != params["path"]]
        if len(new_dotfiles) == len(dotfiles):
            raise DotfileNotFound

        new_dotfiles.append({
            "path": params["path"],
            "perm": params["permission"],
            "data": params["data"],
        })
        dotfile_packed = msgpack.packb(new_dotfiles)
        if len(dotfile_packed) > MAXIMUM_DOTFILE_SIZE:
            raise DotfileCreationFailed("No leftover space for dotfile storage")

        query = (
            domains.update()
            .values(dotfiles=dotfile_packed)
            .where(domains.c.name == params["domain"])
        )
        await conn.execute(query)
    return web.json_response({})


@server_status_required(READ_ALLOWED)
@admin_required
@check_api_params(
    t.Dict({
        t.Key("domain"): t.String,
        t.Key("path"): t.String,
    }),
)
async def delete(request: web.Request, params: Any) -> web.Response:
    log.info("DOMAINCONFIG.DELETE_DOTFILE (domain:{0})", params["domain"])
    if not request["is_superadmin"] and request["user"]["domain_name"] != params["domain"]:
        raise GenericForbidden("Domain admins cannot delete dotfiles of other domains")
    root_ctx: RootContext = request.app["_root.context"]
    async with root_ctx.db.begin() as conn:
        dotfiles, _ = await query_domain_dotfiles(conn, params["domain"])
        if dotfiles is None:
            raise DomainNotFound
        new_dotfiles = [x for x in dotfiles if x["path"] != params["path"]]
        if len(new_dotfiles) == len(dotfiles):
            raise DotfileNotFound

        dotfile_packed = msgpack.packb(new_dotfiles)
        query = (
            domains.update()
            .values(dotfiles=dotfile_packed)
            .where(domains.c.name == params["domain"])
        )
        await conn.execute(query)
        return web.json_response({"success": True})


async def init(app: web.Application) -> None:
    pass


async def shutdown(app: web.Application) -> None:
    pass


def create_app(
    default_cors_options: CORSOptions,
) -> Tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    app["api_versions"] = (4, 5)
    app["prefix"] = "domain-config"
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    cors.add(app.router.add_route("POST", "/dotfiles", create))
    cors.add(app.router.add_route("GET", "/dotfiles", list_or_get))
    cors.add(app.router.add_route("PATCH", "/dotfiles", update))
    cors.add(app.router.add_route("DELETE", "/dotfiles", delete))

    return app, []



import logging
from datetime import datetime

from ai.backend.common.events.event_types.image.anycast import (
    ImagePullFailedEvent,
    ImagePullFinishedEvent,
    ImagePullStartedEvent,
)
from ai.backend.common.types import (
    AgentId,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.registry import AgentRegistry

from ...models.utils import (
    ExtendedAsyncSAEngine,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class ImageEventHandler:
    def __init__(self, registry: AgentRegistry, db: ExtendedAsyncSAEngine) -> None:
        self._registry = registry
        self._db = db

    async def handle_image_pull_started(
        self,
        context: None,
        agent_id: AgentId,
        ev: ImagePullStartedEvent,
    ) -> None:
        dt = datetime.fromtimestamp(ev.timestamp)
        log.debug("handle_image_pull_started: ag:{} img:{}, start_dt:{}", ev.agent_id, ev.image, dt)
        async with self._db.connect() as db_conn:
            await self._registry.mark_image_pull_started(
                ev.agent_id, ev.image, ev.image_ref, db_conn=db_conn
            )

    async def handle_image_pull_finished(
        self, context: None, agent_id: AgentId, ev: ImagePullFinishedEvent
    ) -> None:
        dt = datetime.fromtimestamp(ev.timestamp)
        log.debug("handle_image_pull_finished: ag:{} img:{}, end_dt:{}", ev.agent_id, ev.image, dt)
        async with self._db.connect() as db_conn:
            await self._registry.mark_image_pull_finished(
                ev.agent_id, ev.image, ev.image_ref, db_conn=db_conn
            )

    async def handle_image_pull_failed(
        self,
        context: None,
        agent_id: AgentId,
        ev: ImagePullFailedEvent,
    ) -> None:
        log.warning("handle_image_pull_failed: ag:{} img:{}, msg:{}", ev.agent_id, ev.image, ev.msg)
        async with self._db.connect() as db_conn:
            await self._registry.handle_image_pull_failed(
                ev.agent_id, ev.image, ev.msg, ev.image_ref, db_conn=db_conn
            )



import asyncio
import logging
import secrets
import uuid
from http import HTTPStatus
from typing import Awaitable, Callable

import aiohttp
import sqlalchemy as sa
import tomli
from sqlalchemy.orm import selectinload
from sqlalchemy.orm.exc import NoResultFound
from yarl import URL

from ai.backend.common import redis_helper
from ai.backend.common.bgtask.bgtask import BackgroundTaskManager, ProgressReporter
from ai.backend.common.events.dispatcher import (
    EventDispatcher,
    EventHandler,
)
from ai.backend.common.events.event_types.kernel.types import (
    KernelLifecycleEventReason,
)
from ai.backend.common.events.event_types.model_serving.broadcast import (
    ModelServiceStatusBroadcastEvent,
)
from ai.backend.common.events.event_types.session.broadcast import (
    SessionCancelledBroadcastEvent,
    SessionEnqueuedBroadcastEvent,
    SessionPreparingBroadcastEvent,
    SessionStartedBroadcastEvent,
    SessionTerminatedBroadcastEvent,
)
from ai.backend.common.json import dump_json_str
from ai.backend.common.types import (
    AgentId,
    ClusterMode,
    ImageAlias,
    MountPermission,
    MountTypes,
    RedisConnectionInfo,
    RuntimeVariant,
    SessionTypes,
)
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.config.constant import DEFAULT_CHUNK_SIZE
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.models.endpoint import (
    EndpointLifecycle,
    EndpointRow,
    EndpointTokenRow,
    ModelServiceHelper,
)
from ai.backend.manager.models.group import resolve_group_name_or_id
from ai.backend.manager.models.image import ImageIdentifier, ImageRow
from ai.backend.manager.models.keypair import KeyPairRow
from ai.backend.manager.models.resource_policy import keypair_resource_policies
from ai.backend.manager.models.routing import RouteStatus, RoutingRow
from ai.backend.manager.models.scaling_group import scaling_groups
from ai.backend.manager.models.session import KernelLoadingStrategy, SessionRow
from ai.backend.manager.models.storage import StorageSessionManager
from ai.backend.manager.models.user import UserRole, UserRow
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine, execute_with_retry
from ai.backend.manager.models.vfolder import VFolderOwnershipType, VFolderRow
from ai.backend.manager.registry import AgentRegistry
from ai.backend.manager.services.model_serving.actions.clear_error import (
    ClearErrorAction,
    ClearErrorActionResult,
)
from ai.backend.manager.services.model_serving.actions.create_model_service import (
    CreateModelServiceAction,
    CreateModelServiceActionResult,
)
from ai.backend.manager.services.model_serving.actions.delete_model_service import (
    DeleteModelServiceAction,
    DeleteModelServiceActionResult,
)
from ai.backend.manager.services.model_serving.actions.delete_route import (
    DeleteRouteAction,
    DeleteRouteActionResult,
)
from ai.backend.manager.services.model_serving.actions.dry_run_model_service import (
    DryRunModelServiceAction,
    DryRunModelServiceActionResult,
)
from ai.backend.manager.services.model_serving.actions.force_sync import (
    ForceSyncAction,
    ForceSyncActionResult,
)
from ai.backend.manager.services.model_serving.actions.generate_token import (
    GenerateTokenAction,
    GenerateTokenActionResult,
)
from ai.backend.manager.services.model_serving.actions.get_model_service_info import (
    GetModelServiceInfoAction,
    GetModelServiceInfoActionResult,
)
from ai.backend.manager.services.model_serving.actions.list_errors import (
    ListErrorsAction,
    ListErrorsActionResult,
)
from ai.backend.manager.services.model_serving.actions.list_model_service import (
    ListModelServiceAction,
    ListModelServiceActionResult,
)
from ai.backend.manager.services.model_serving.actions.modify_endpoint import (
    ModifyEndpointAction,
    ModifyEndpointActionResult,
)
from ai.backend.manager.services.model_serving.actions.update_route import (
    UpdateRouteAction,
    UpdateRouteActionResult,
)
from ai.backend.manager.services.model_serving.exceptions import (
    EndpointNotFound,
    InvalidAPIParameters,
    ModelServiceNotFound,
    RouteNotFound,
)
from ai.backend.manager.services.model_serving.services.utils import verify_user_access_scopes
from ai.backend.manager.services.model_serving.types import (
    CompactServiceInfo,
    EndpointData,
    ErrorInfo,
    ModelServiceDefinition,
    MutationResult,
    RouteInfo,
    ServiceInfo,
)
from ai.backend.manager.types import MountOptionModel, UserScope

log = BraceStyleAdapter(logging.getLogger(__name__))


class ModelServingService:
    _db: ExtendedAsyncSAEngine
    _agent_registry: AgentRegistry
    _background_task_manager: BackgroundTaskManager
    _event_dispatcher: EventDispatcher
    _storage_manager: StorageSessionManager
    _config_provider: ManagerConfigProvider

    _redis_live: RedisConnectionInfo

    def __init__(
        self,
        db: ExtendedAsyncSAEngine,
        agent_registry: AgentRegistry,
        background_task_manager: BackgroundTaskManager,
        event_dispatcher: EventDispatcher,
        storage_manager: StorageSessionManager,
        config_provider: ManagerConfigProvider,
        redis_live: RedisConnectionInfo,
    ) -> None:
        self._db = db
        self._agent_registry = agent_registry
        self._background_task_manager = background_task_manager
        self._event_dispatcher = event_dispatcher
        self._storage_manager = storage_manager
        self._config_provider = config_provider
        self._redis_live = redis_live

    async def _fetch_file_from_storage_proxy(
        self,
        filename: str,
        model_vfolder_row: VFolderRow,
    ) -> bytes:
        vfid = model_vfolder_row.vfid
        folder_host = model_vfolder_row.host

        proxy_name, volume_name = self._storage_manager.get_proxy_and_volume(folder_host)

        chunks = bytes()
        async with self._storage_manager.request(
            proxy_name,
            "POST",
            "folder/file/fetch",
            json={
                "volume": volume_name,
                "vfid": str(vfid),
                "relpath": f"./{filename}",
            },
        ) as (_, storage_resp):
            while True:
                chunk = await storage_resp.content.read(DEFAULT_CHUNK_SIZE)
                if not chunk:
                    break
                chunks += chunk
        return chunks

    async def create(self, action: CreateModelServiceAction) -> CreateModelServiceActionResult:
        service_prepare_ctx = action.creator.model_service_prepare_ctx

        async with self._db.begin_readonly_session() as db_sess:
            model_vfolder_row = await VFolderRow.get(db_sess, service_prepare_ctx.model_id)
            if model_vfolder_row.ownership_type == VFolderOwnershipType.GROUP:
                raise InvalidAPIParameters(
                    "Cannot create model service with the project type's vfolder"
                )

            chunks = await self._fetch_file_from_storage_proxy(
                "service-definition.toml", model_vfolder_row
            )

            if chunks:
                raw_service_definition = chunks.decode("utf-8")
                service_definition = tomli.loads(raw_service_definition)

                definition = action.creator.runtime_variant
                if definition in service_definition:
                    variant_def = ModelServiceDefinition.model_validate(
                        service_definition[definition]
                    )
                    if variant_def.resource_slots:
                        action.creator.config.resources = variant_def.resource_slots
                    if variant_def.environment:
                        action.creator.image = variant_def.environment.image
                        action.creator.architecture = variant_def.environment.architecture
                    if variant_def.environ:
                        if action.creator.config.environ:
                            action.creator.config.environ.update(variant_def.environ)
                        else:
                            action.creator.config.environ = variant_def.environ

            image_row = await ImageRow.resolve(
                db_sess,
                [
                    ImageIdentifier(action.creator.image, action.creator.architecture),
                    ImageAlias(action.creator.image),
                ],
            )

        creation_config = action.creator.config.to_dict()
        creation_config["mounts"] = [
            service_prepare_ctx.model_id,
            *[m.vfid.folder_id for m in service_prepare_ctx.extra_mounts],
        ]
        creation_config["mount_map"] = {
            service_prepare_ctx.model_id: action.creator.config.model_mount_destination,
            **{
                m.vfid.folder_id: m.kernel_path.as_posix() for m in service_prepare_ctx.extra_mounts
            },
        }
        creation_config["mount_options"] = {
            m.vfid.folder_id: {"permission": m.mount_perm} for m in service_prepare_ctx.extra_mounts
        }
        sudo_session_enabled = action.creator.sudo_session_enabled

        # check if session is valid to be created
        await self._agent_registry.create_session(
            "",
            image_row.image_ref,
            UserScope(
                domain_name=action.creator.domain_name,
                group_id=service_prepare_ctx.group_id,
                user_uuid=service_prepare_ctx.owner_uuid,
                user_role=service_prepare_ctx.owner_role,
            ),
            service_prepare_ctx.owner_access_key,
            service_prepare_ctx.resource_policy,
            SessionTypes.INFERENCE,
            creation_config,
            action.creator.cluster_mode,
            action.creator.cluster_size,
            dry_run=True,  # Setting this to True will prevent actual session from being enqueued
            bootstrap_script=action.creator.bootstrap_script,
            startup_command=action.creator.startup_command,
            tag=action.creator.tag,
            callback_url=URL(action.creator.callback_url.unicode_string())
            if action.creator.callback_url
            else None,
            sudo_session_enabled=sudo_session_enabled,
        )

        async with self._db.begin_session() as db_sess:
            query = sa.select(EndpointRow).where(
                (EndpointRow.lifecycle_stage != EndpointLifecycle.DESTROYED)
                & (EndpointRow.name == action.creator.service_name)
            )
            result = await db_sess.execute(query)
            service_with_duplicate_name = result.scalar()
            if service_with_duplicate_name is not None:
                raise InvalidAPIParameters("Cannot create multiple services with same name")

            project_id = await resolve_group_name_or_id(
                await db_sess.connection(), action.creator.domain_name, action.creator.group_name
            )
            if project_id is None:
                raise InvalidAPIParameters(f"Invalid group name {project_id}")
            endpoint = EndpointRow(
                action.creator.service_name,
                service_prepare_ctx.model_definition_path,
                action.request_user_id,
                service_prepare_ctx.owner_uuid,
                action.creator.replicas,
                image_row,
                service_prepare_ctx.model_id,
                action.creator.domain_name,
                project_id,
                service_prepare_ctx.scaling_group,
                action.creator.config.resources,
                action.creator.cluster_mode,
                action.creator.cluster_size,
                service_prepare_ctx.extra_mounts,
                model_mount_destination=action.creator.config.model_mount_destination,
                tag=action.creator.tag,
                startup_command=action.creator.startup_command,
                callback_url=URL(action.creator.callback_url.unicode_string())
                if action.creator.callback_url
                else None,
                environ=action.creator.config.environ,
                bootstrap_script=action.creator.bootstrap_script,
                resource_opts=action.creator.config.resource_opts,
                open_to_public=action.creator.open_to_public,
                runtime_variant=action.creator.runtime_variant,
            )
            endpoint.url = await self._agent_registry.create_appproxy_endpoint(db_sess, endpoint)
            db_sess.add(endpoint)
            await db_sess.flush()
            endpoint_id = endpoint.id

        return CreateModelServiceActionResult(
            ServiceInfo(
                endpoint_id=endpoint_id,
                model_id=endpoint.model,
                extra_mounts=[m.vfid.folder_id for m in endpoint.extra_mounts],
                name=action.creator.service_name,
                model_definition_path=service_prepare_ctx.model_definition_path,
                replicas=endpoint.replicas,
                desired_session_count=endpoint.replicas,
                active_routes=[],
                service_endpoint=None,
                is_public=action.creator.open_to_public,
                runtime_variant=action.creator.runtime_variant,
            )
        )

    async def list_serve(self, action: ListModelServiceAction) -> ListModelServiceActionResult:
        query_conds = (EndpointRow.session_owner == action.session_owener_id) & (
            EndpointRow.lifecycle_stage == EndpointLifecycle.CREATED
        )
        if action.name:
            query_conds &= EndpointRow.name == action.name

        async with self._db.begin_readonly_session() as db_sess:
            query = (
                sa.select(EndpointRow)
                .where(query_conds)
                .options(selectinload(EndpointRow.routings))
            )
            result = await db_sess.execute(query)
            rows = result.scalars().all()

        return ListModelServiceActionResult(
            data=[
                CompactServiceInfo(
                    id=endpoint.id,
                    name=endpoint.name,
                    replicas=endpoint.replicas,
                    desired_session_count=endpoint.replicas,
                    active_route_count=len([
                        r for r in endpoint.routings if r.status == RouteStatus.HEALTHY
                    ]),
                    service_endpoint=endpoint.url,
                    is_public=endpoint.open_to_public,
                )
                for endpoint in rows
            ]
        )

    async def delete(self, action: DeleteModelServiceAction) -> DeleteModelServiceActionResult:
        service_id = action.service_id
        async with self._db.begin_readonly_session() as db_sess:
            try:
                endpoint = await EndpointRow.get(db_sess, service_id, load_routes=True)
            except NoResultFound:
                raise ModelServiceNotFound
        await verify_user_access_scopes(self._db, action.requester_ctx, endpoint.session_owner)

        async with self._db.begin_session() as db_sess:
            if len(endpoint.routings) == 0:
                query = (
                    sa.update(EndpointRow)
                    .where(EndpointRow.id == service_id)
                    .values({
                        "lifecycle_stage": EndpointLifecycle.DESTROYED,
                        "destroyed_at": sa.func.now(),
                    })
                )
            else:
                query = (
                    sa.update(EndpointRow)
                    .where(EndpointRow.id == service_id)
                    .values({
                        "replicas": 0,
                        "lifecycle_stage": EndpointLifecycle.DESTROYING,
                    })
                )
            await db_sess.execute(query)

        return DeleteModelServiceActionResult(success=True)

    async def dry_run(self, action: DryRunModelServiceAction) -> DryRunModelServiceActionResult:
        # TODO: Seperate background task definition and trigger into different layer
        service_prepare_ctx = action.model_service_prepare_ctx
        async with self._db.begin_readonly_session() as session:
            image_row = await ImageRow.resolve(
                session,
                [
                    ImageIdentifier(action.image, action.architecture),
                    ImageAlias(action.image),
                ],
            )
            query = sa.select(sa.join(UserRow, KeyPairRow, KeyPairRow.user == UserRow.uuid)).where(
                UserRow.uuid == action.request_user_id
            )
            created_user = (await session.execute(query)).fetchone()

        creation_config = action.config.to_dict()
        creation_config["mount_map"] = {
            service_prepare_ctx.model_id: action.config.model_mount_destination
        }
        sudo_session_enabled = action.sudo_session_enabled

        async def _task(reporter: ProgressReporter) -> None:
            terminated_event = asyncio.Event()

            result = await self._agent_registry.create_session(
                f"model-eval-{secrets.token_urlsafe(16)}",
                image_row.image_ref,
                UserScope(
                    domain_name=action.domain_name,
                    group_id=service_prepare_ctx.group_id,
                    user_uuid=created_user.uuid,
                    user_role=created_user.role,
                ),
                service_prepare_ctx.owner_access_key,
                service_prepare_ctx.resource_policy,
                SessionTypes.INFERENCE,
                {
                    "mounts": [
                        service_prepare_ctx.model_id,
                        *[m.vfid for m in service_prepare_ctx.extra_mounts],
                    ],
                    "mount_map": {
                        service_prepare_ctx.model_id: action.config.model_mount_destination,
                        **{m.vfid: m.kernel_path for m in service_prepare_ctx.extra_mounts},
                    },
                    "mount_options": {
                        m.vfid: {"permission": m.mount_perm}
                        for m in service_prepare_ctx.extra_mounts
                    },
                    "model_definition_path": service_prepare_ctx.model_definition_path,
                    "environ": creation_config["environ"],
                    "scaling_group": service_prepare_ctx.scaling_group,
                    "resources": creation_config["resources"],
                    "resource_opts": creation_config["resource_opts"],
                    "preopen_ports": None,
                    "agent_list": None,
                },
                action.cluster_mode,
                action.cluster_size,
                bootstrap_script=action.bootstrap_script,
                startup_command=action.startup_command,
                tag=action.tag,
                callback_url=URL(action.callback_url.unicode_string())
                if action.callback_url
                else None,
                enqueue_only=True,
                sudo_session_enabled=sudo_session_enabled,
            )

            await reporter.update(
                message=dump_json_str({
                    "event": "session_enqueued",
                    "session_id": str(result["sessionId"]),
                })
            )

            async def _handle_event(
                context: None,
                source: AgentId,
                event: SessionEnqueuedBroadcastEvent
                | SessionPreparingBroadcastEvent
                | SessionStartedBroadcastEvent
                | SessionCancelledBroadcastEvent
                | SessionTerminatedBroadcastEvent
                | ModelServiceStatusBroadcastEvent,
            ) -> None:
                task_message = {"event": event.event_name(), "session_id": str(event.session_id)}
                match event:
                    case ModelServiceStatusBroadcastEvent():
                        task_message["is_healthy"] = event.new_status.value
                await reporter.update(message=dump_json_str(task_message))

                match event:
                    case SessionTerminatedBroadcastEvent() | SessionCancelledBroadcastEvent():
                        terminated_event.set()
                    case ModelServiceStatusBroadcastEvent():
                        async with self._db.begin_readonly_session() as db_sess:
                            session = await SessionRow.get_session(
                                db_sess,
                                result["sessionId"],
                                None,
                                kernel_loading_strategy=KernelLoadingStrategy.ALL_KERNELS,
                            )
                            await self._agent_registry.destroy_session(
                                session,
                                forced=True,
                            )

            session_event_matcher = lambda args: args[0] == str(result["sessionId"])
            model_service_event_matcher = lambda args: args[1] == str(result["sessionId"])

            handlers: list[EventHandler] = [
                self._event_dispatcher.subscribe(
                    SessionPreparingBroadcastEvent,
                    None,
                    _handle_event,
                    args_matcher=session_event_matcher,
                ),
                self._event_dispatcher.subscribe(
                    SessionStartedBroadcastEvent,
                    None,
                    _handle_event,
                    args_matcher=session_event_matcher,
                ),
                self._event_dispatcher.subscribe(
                    SessionCancelledBroadcastEvent,
                    None,
                    _handle_event,
                    args_matcher=session_event_matcher,
                ),
                self._event_dispatcher.subscribe(
                    SessionTerminatedBroadcastEvent,
                    None,
                    _handle_event,
                    args_matcher=session_event_matcher,
                ),
                self._event_dispatcher.subscribe(
                    ModelServiceStatusBroadcastEvent,
                    None,
                    _handle_event,
                    args_matcher=model_service_event_matcher,
                ),
            ]

            try:
                await terminated_event.wait()
            finally:
                for handler in handlers:
                    self._event_dispatcher.unsubscribe(handler)

        task_id = await self._background_task_manager.start(_task)
        return DryRunModelServiceActionResult(task_id)

    async def get_model_service_info(
        self, action: GetModelServiceInfoAction
    ) -> GetModelServiceInfoActionResult:
        async with self._db.begin_readonly_session() as db_sess:
            try:
                endpoint = await EndpointRow.get(db_sess, action.service_id, load_routes=True)
            except NoResultFound:
                raise ModelServiceNotFound

        await verify_user_access_scopes(self._db, action.requester_ctx, endpoint.session_owner)
        return GetModelServiceInfoActionResult(
            ServiceInfo(
                endpoint_id=endpoint.id,
                model_id=endpoint.model,
                extra_mounts=[m.vfid.folder_id for m in endpoint.extra_mounts],
                name=endpoint.name,
                model_definition_path=endpoint.model_definition_path,
                replicas=endpoint.replicas,
                desired_session_count=endpoint.replicas,
                active_routes=[
                    RouteInfo(
                        route_id=r.id,
                        session_id=r.session,
                        traffic_ratio=r.traffic_ratio,
                    )
                    for r in endpoint.routings
                ],
                service_endpoint=endpoint.url,
                is_public=endpoint.open_to_public,
                runtime_variant=endpoint.runtime_variant,
            )
        )

    async def list_errors(self, action: ListErrorsAction) -> ListErrorsActionResult:
        async with self._db.begin_readonly_session() as db_sess:
            try:
                endpoint = await EndpointRow.get(db_sess, action.service_id, load_routes=True)
            except NoResultFound:
                raise ModelServiceNotFound

        await verify_user_access_scopes(self._db, action.requester_ctx, endpoint.session_owner)

        error_routes = [r for r in endpoint.routings if r.status == RouteStatus.FAILED_TO_START]

        return ListErrorsActionResult(
            error_info=[
                ErrorInfo(
                    session_id=route.error_data.get("session_id"), error=route.error_data["errors"]
                )
                for route in error_routes
            ],
            retries=endpoint.retries,
        )

    async def clear_error(self, action: ClearErrorAction) -> ClearErrorActionResult:
        async with self._db.begin_readonly_session() as db_sess:
            try:
                endpoint = await EndpointRow.get(db_sess, action.service_id, load_routes=True)
            except NoResultFound:
                raise ModelServiceNotFound
        await verify_user_access_scopes(self._db, action.requester_ctx, endpoint.session_owner)

        async with self._db.begin_session() as db_sess:
            query = sa.delete(RoutingRow).where(
                (RoutingRow.endpoint == action.service_id)
                & (RoutingRow.status == RouteStatus.FAILED_TO_START)
            )
            await db_sess.execute(query)
            query = (
                sa.update(EndpointRow).values({"retries": 0}).where(EndpointRow.id == endpoint.id)
            )
            await db_sess.execute(query)

        return ClearErrorActionResult(success=True)

    async def update_route(self, action: UpdateRouteAction) -> UpdateRouteActionResult:
        async with self._db.begin_session() as db_sess:
            try:
                route = await RoutingRow.get(db_sess, action.route_id, load_endpoint=True)
            except NoResultFound:
                raise RouteNotFound
            if route.endpoint != action.service_id:
                raise RouteNotFound
            await verify_user_access_scopes(
                self._db, action.requester_ctx, route.endpoint_row.session_owner
            )

            route.traffic_ratio = action.traffic_ratio
            await redis_helper.execute(
                self._redis_live,
                lambda r: r.set(
                    f"endpoint.{action.service_id}.session.{route.session}.traffic_ratio",
                    str(action.traffic_ratio),
                    ex=3600,
                ),
            )

            await self._agent_registry.notify_endpoint_route_update_to_appproxy(route.endpoint_row)
            await db_sess.commit()

        return UpdateRouteActionResult(success=True)

    async def delete_route(self, action: DeleteRouteAction) -> DeleteRouteActionResult:
        async with self._db.begin_readonly_session() as db_sess:
            try:
                route = await RoutingRow.get(
                    db_sess, action.route_id, load_endpoint=True, load_session=True
                )
            except NoResultFound:
                raise RouteNotFound
            if route.endpoint != action.service_id:
                raise RouteNotFound
        await verify_user_access_scopes(
            self._db, action.requester_ctx, route.endpoint_row.session_owner
        )
        if route.status == RouteStatus.PROVISIONING:
            raise InvalidAPIParameters("Cannot remove route in PROVISIONING status")

        await self._agent_registry.destroy_session(
            route.session_row,
            forced=False,
            reason=KernelLifecycleEventReason.SERVICE_SCALED_DOWN,
        )

        async with self._db.begin_session() as db_sess:
            query = (
                sa.update(EndpointRow)
                .where(EndpointRow.id == action.service_id)
                .values({"replicas": route.endpoint_row.replicas - 1})
            )
            await db_sess.execute(query)

        return DeleteRouteActionResult(success=True)

    async def generate_token(self, action: GenerateTokenAction) -> GenerateTokenActionResult:
        async with self._db.begin_readonly_session() as db_sess:
            try:
                endpoint = await EndpointRow.get(db_sess, action.service_id, load_routes=True)
            except NoResultFound:
                raise ModelServiceNotFound
            query = (
                sa.select([scaling_groups.c.wsproxy_addr, scaling_groups.c.wsproxy_api_token])
                .select_from(scaling_groups)
                .where((scaling_groups.c.name == endpoint.resource_group))
            )

            result = await db_sess.execute(query)
            sgroup = result.first()
            wsproxy_addr = sgroup["wsproxy_addr"]
            wsproxy_api_token = sgroup["wsproxy_api_token"]

        await verify_user_access_scopes(self._db, action.requester_ctx, endpoint.session_owner)
        body = {"user_uuid": str(endpoint.session_owner), "exp": action.expires_at}
        async with aiohttp.ClientSession() as session:
            async with session.post(
                f"{wsproxy_addr}/v2/endpoints/{endpoint.id}/token",
                json=body,
                headers={
                    "accept": "application/json",
                    "X-BackendAI-Token": wsproxy_api_token,
                },
            ) as resp:
                resp_json = await resp.json()
                if resp.status != HTTPStatus.OK:
                    raise EndpointNotFound(
                        f"Failed to generate token: {resp.status} {resp.reason} {resp_json}"
                    )
                token = resp_json["token"]

        async with self._db.begin_session() as db_sess:
            token_id = uuid.uuid4()
            token_row = EndpointTokenRow(
                token_id,
                token,
                endpoint.id,
                endpoint.domain,
                endpoint.project,
                endpoint.session_owner,
            )
            db_sess.add(token_row)
            await db_sess.commit()
            await db_sess.refresh(token_row)

        return GenerateTokenActionResult(token_row.to_dataclass())

    async def force_sync_with_app_proxy(self, action: ForceSyncAction) -> ForceSyncActionResult:
        async with self._db.begin_readonly_session() as db_sess:
            try:
                endpoint = await EndpointRow.get(db_sess, action.service_id, load_routes=True)
            except NoResultFound:
                raise ModelServiceNotFound
        await verify_user_access_scopes(self._db, action.requester_ctx, endpoint.session_owner)

        await self._agent_registry.notify_endpoint_route_update_to_appproxy(endpoint)

        return ForceSyncActionResult(success=True)

    async def modify_endpoint(self, action: ModifyEndpointAction) -> ModifyEndpointActionResult:
        async def _do_mutate() -> MutationResult:
            async with self._db.begin_session() as db_session:
                try:
                    endpoint_row = await EndpointRow.get(
                        db_session,
                        action.endpoint_id,
                        load_session_owner=True,
                        load_model=True,
                        load_routes=True,
                    )
                    match action.requester_ctx.user_role:
                        case UserRole.SUPERADMIN:
                            pass
                        case UserRole.ADMIN:
                            domain_name = action.requester_ctx.domain_name
                            if endpoint_row.domain != domain_name:
                                raise EndpointNotFound
                        case _:
                            user_id = action.requester_ctx.user_id
                            if endpoint_row.session_owner != user_id:
                                raise EndpointNotFound
                except NoResultFound:
                    raise EndpointNotFound
                if endpoint_row.lifecycle_stage in (
                    EndpointLifecycle.DESTROYING,
                    EndpointLifecycle.DESTROYED,
                ):
                    raise InvalidAPIParameters("Cannot update endpoint marked for removal")

                fields_to_update = action.modifier.fields_to_update()
                for key, value in fields_to_update.items():
                    setattr(endpoint_row, key, value)

                fields_to_update_require_none_check = (
                    action.modifier.fields_to_update_require_none_check()
                )
                for key, value in fields_to_update_require_none_check.items():
                    if value is not None:
                        setattr(endpoint_row, key, value)

                image_ref = action.modifier.image.optional_value()
                if image_ref is not None:
                    image_name = image_ref.name
                    arch = image_ref.architecture.value()
                    image_row = await ImageRow.resolve(
                        db_session, [ImageIdentifier(image_name, arch), ImageAlias(image_name)]
                    )
                    endpoint_row.image = image_row.id

                session_owner: UserRow = endpoint_row.session_owner_row

                conn = await db_session.connection()
                assert conn

                await ModelServiceHelper.check_scaling_group(
                    conn,
                    endpoint_row.resource_group,
                    session_owner.main_access_key,
                    endpoint_row.domain,
                    endpoint_row.project,
                )

                user_scope = UserScope(
                    domain_name=endpoint_row.domain,
                    group_id=endpoint_row.project,
                    user_uuid=session_owner.uuid,
                    user_role=session_owner.role,
                )

                query = (
                    sa.select([keypair_resource_policies])
                    .select_from(keypair_resource_policies)
                    .where(keypair_resource_policies.c.name == session_owner.resource_policy)
                )
                result = await conn.execute(query)

                resource_policy = result.first()
                extra_mounts_input = action.modifier.extra_mounts.optional_value()
                if extra_mounts_input is not None:
                    extra_mounts = {
                        mount.vfolder_id.value(): MountOptionModel(
                            mount_destination=(
                                mount.mount_destination.value()
                                if mount.mount_destination.optional_value() is not None
                                else None
                            ),
                            type=MountTypes(mount.type.value())
                            if mount.type.optional_value() is not None
                            else MountTypes.BIND,
                            permission=(
                                MountPermission(mount.permission.value())
                                if mount.permission.optional_value() is not None
                                else None
                            ),
                        )
                        for mount in extra_mounts_input
                    }
                    vfolder_mounts = await ModelServiceHelper.check_extra_mounts(
                        conn,
                        self._config_provider.legacy_etcd_config_loader,
                        self._storage_manager,
                        endpoint_row.model,
                        endpoint_row.model_mount_destination,
                        extra_mounts,
                        user_scope,
                        resource_policy,
                    )
                    endpoint_row.extra_mounts = vfolder_mounts

                if endpoint_row.runtime_variant == RuntimeVariant.CUSTOM:
                    model_definition_path = (
                        await ModelServiceHelper.validate_model_definition_file_exists(
                            self._storage_manager,
                            endpoint_row.model_row.host,
                            endpoint_row.model_row.vfid,
                            endpoint_row.model_definition_path,
                        )
                    )
                    await ModelServiceHelper.validate_model_definition(
                        self._storage_manager,
                        endpoint_row.model_row.host,
                        endpoint_row.model_row.vfid,
                        model_definition_path,
                    )
                elif (
                    endpoint_row.runtime_variant != RuntimeVariant.CMD
                    and endpoint_row.model_mount_destination != "/models"
                ):
                    raise InvalidAPIParameters(
                        "Model mount destination must be /models for non-custom runtimes"
                    )

                async with self._db.begin_session() as db_session:
                    image_row = await ImageRow.resolve(
                        db_session,
                        [
                            ImageIdentifier(
                                endpoint_row.image_row.name, endpoint_row.image_row.architecture
                            ),
                        ],
                    )

                await self._agent_registry.create_session(
                    "",
                    image_row.image_ref,
                    user_scope,
                    session_owner.main_access_key,
                    resource_policy,
                    SessionTypes.INFERENCE,
                    {
                        "mounts": [
                            endpoint_row.model,
                            *[m.vfid.folder_id for m in endpoint_row.extra_mounts],
                        ],
                        "mount_map": {
                            endpoint_row.model: endpoint_row.model_mount_destination,
                            **{
                                m.vfid.folder_id: m.kernel_path.as_posix()
                                for m in endpoint_row.extra_mounts
                            },
                        },
                        "mount_options": {
                            m.vfid.folder_id: {"permission": m.mount_perm}
                            for m in endpoint_row.extra_mounts
                        },
                        "environ": endpoint_row.environ,
                        "scaling_group": endpoint_row.resource_group,
                        "resources": endpoint_row.resource_slots,
                        "resource_opts": endpoint_row.resource_opts,
                        "preopen_ports": None,
                        "agent_list": None,
                    },
                    ClusterMode(endpoint_row.cluster_mode),
                    endpoint_row.cluster_size,
                    bootstrap_script=endpoint_row.bootstrap_script,
                    startup_command=endpoint_row.startup_command,
                    tag=endpoint_row.tag,
                    callback_url=endpoint_row.callback_url,
                    sudo_session_enabled=session_owner.sudo_session_enabled,
                    dry_run=True,
                )

                await db_session.commit()
                return MutationResult(
                    success=True,
                    message="success",
                    data=EndpointData.from_row(endpoint_row),
                )

        result = await self._db_mutation_wrapper(_do_mutate)
        return ModifyEndpointActionResult(success=result.success, data=result.data)

    async def _db_mutation_wrapper(
        self, _do_mutate: Callable[[], Awaitable[MutationResult]]
    ) -> MutationResult:
        try:
            return await execute_with_retry(_do_mutate)
        except sa.exc.IntegrityError as e:
            log.warning("db_mutation_wrapper(): integrity error ({})", repr(e))
            return MutationResult(success=False, message=f"integrity error: {e}", data=None)
        except sa.exc.StatementError as e:
            log.warning(
                "db_mutation_wrapper(): statement error ({})\n{}",
                repr(e),
                e.statement or "(unknown)",
            )
            orig_exc = e.orig
            return MutationResult(success=False, message=str(orig_exc), data=None)
        except (asyncio.CancelledError, asyncio.TimeoutError):
            raise
        except Exception:
            log.exception("db_mutation_wrapper(): other error")
            raise



import logging

from ai.backend.common.events.event_types.vfolder.anycast import (
    VFolderDeletionFailureEvent,
    VFolderDeletionSuccessEvent,
)
from ai.backend.common.types import (
    AgentId,
)
from ai.backend.logging import BraceStyleAdapter

from ...models.utils import (
    ExtendedAsyncSAEngine,
)
from ...models.vfolder import VFolderOperationStatus, update_vfolder_status

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class VFolderEventHandler:
    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db

    async def handle_vfolder_deletion_success(
        self,
        context: None,
        source: AgentId,
        event: VFolderDeletionSuccessEvent,
    ) -> None:
        await update_vfolder_status(
            self._db, [event.vfid.folder_id], VFolderOperationStatus.DELETE_COMPLETE, do_log=True
        )

    async def handle_vfolder_deletion_failure(
        self,
        context: None,
        source: AgentId,
        event: VFolderDeletionFailureEvent,
    ) -> None:
        log.exception(f"Failed to delete vfolder (vfid:{event.vfid}, msg:{event.message})")
        await update_vfolder_status(
            self._db, [event.vfid.folder_id], VFolderOperationStatus.DELETE_ERROR, do_log=True
        )



from dataclasses import dataclass
from typing import Optional, override

from aiohttp import web

from ai.backend.common.dto.manager.auth.field import AuthTokenType
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.config.unified import AuthConfig
from ai.backend.manager.data.auth.types import AuthorizationResult
from ai.backend.manager.services.auth.actions.base import AuthAction


@dataclass
class AuthorizeAction(AuthAction):
    request: web.Request
    type: AuthTokenType
    domain_name: str
    email: str
    password: str
    auth_config: Optional[AuthConfig]

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "authorize"

    @property
    def hook_params(self) -> dict[str, str]:
        return {
            "type": self.type.value,
            "domain": self.domain_name,
            "username": self.email,
            "password": self.password,
        }


@dataclass
class AuthorizeActionResult(BaseActionResult):
    stream_response: Optional[web.StreamResponse]
    authorization_result: Optional[AuthorizationResult]

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.authorization_result.user_id) if self.authorization_result else None



from ai.backend.common.events.dispatcher import AbstractEvent
from ai.backend.common.events.hub.hub import EventHub
from ai.backend.common.types import AgentId
from ai.backend.manager.errors.exceptions import InternalServerError


class PropagatorEventHandler:
    _event_hub: EventHub

    def __init__(self, event_hub: EventHub) -> None:
        self._event_hub = event_hub

    async def propagate_event(
        self,
        context: None,
        source: AgentId,
        event: AbstractEvent,
    ) -> None:
        await self._event_hub.propagate_event(event)

    async def propagate_event_with_close(
        self,
        context: None,
        source: AgentId,
        event: AbstractEvent,
    ) -> None:
        await self._event_hub.propagate_event(event)
        domain_id = event.domain_id()
        if domain_id is None:
            raise InternalServerError("Event domain ID is None")
        await self._event_hub.close_by_alias(event.event_domain(), domain_id)



import uuid
from dataclasses import dataclass, field
from datetime import datetime
from decimal import Decimal
from typing import Any, Mapping, Optional, Self, Sequence, override

import yarl
from pydantic import AnyUrl, BaseModel, Field, HttpUrl

from ai.backend.common.types import (
    AccessKey,
    AutoScalingMetricComparator,
    AutoScalingMetricSource,
    ClusterMode,
    MountPermission,
    MountTypes,
    ResourceSlot,
    RuntimeVariant,
    VFolderMount,
)
from ai.backend.manager.data.image.types import ImageData
from ai.backend.manager.models.endpoint import EndpointAutoScalingRuleRow, EndpointRow
from ai.backend.manager.models.routing import RouteStatus, RoutingRow
from ai.backend.manager.models.user import UserRole
from ai.backend.manager.types import Creator, OptionalState, PartialModifier, TriState


@dataclass
class ModelServicePrepareCtx:
    model_id: uuid.UUID
    model_definition_path: Optional[str]
    requester_access_key: AccessKey
    owner_access_key: AccessKey
    owner_uuid: uuid.UUID
    owner_role: UserRole
    group_id: uuid.UUID
    resource_policy: dict
    scaling_group: str
    extra_mounts: Sequence[VFolderMount]


@dataclass
class MountOption:
    mount_destination: Optional[str]
    type: MountTypes
    permission: Optional[MountPermission]

    def to_dict(self) -> dict[str, Any]:
        return {
            "mount_destination": self.mount_destination,
            "type": self.type.value,
            "permission": self.permission.value if self.permission else None,
        }


@dataclass
class RouteInfo:
    route_id: uuid.UUID
    session_id: Optional[uuid.UUID]
    traffic_ratio: float


@dataclass
class RouteConnectionInfo:
    app: str
    kernel_host: str
    kernel_port: int


@dataclass
class ServiceConfig:
    model: str
    model_definition_path: Optional[str]
    model_version: int
    model_mount_destination: str
    extra_mounts: dict[uuid.UUID, MountOption]
    environ: Optional[dict[str, str]]
    scaling_group: str
    resources: dict[str, str | int]
    resource_opts: dict[str, str | int | bool]

    def to_dict(self) -> dict[str, Any]:
        return {
            "model": self.model,
            "model_definition_path": self.model_definition_path,
            "model_mount_destination": self.model_mount_destination,
            "extra_mounts": {key: value.to_dict() for key, value in self.extra_mounts.items()},
            "environ": self.environ if self.environ is not None else {},
            "scaling_group": self.scaling_group,
            "resources": self.resources,
            "resource_opts": self.resource_opts,
        }


@dataclass
class ServiceInfo:
    endpoint_id: uuid.UUID
    model_id: uuid.UUID
    extra_mounts: Sequence[uuid.UUID]
    name: str
    model_definition_path: Optional[str]
    replicas: int
    desired_session_count: int
    active_routes: list[RouteInfo]
    service_endpoint: Optional[HttpUrl]
    is_public: bool
    runtime_variant: RuntimeVariant


@dataclass
class CompactServiceInfo:
    id: uuid.UUID
    name: str
    replicas: int
    desired_session_count: int
    active_route_count: int
    service_endpoint: Optional[HttpUrl]
    is_public: bool


@dataclass
class RequesterCtx:
    is_authorized: Optional[bool]
    user_id: uuid.UUID
    user_role: UserRole
    domain_name: str


@dataclass
class ErrorInfo:
    session_id: Optional[uuid.UUID]
    error: dict[str, Any]


@dataclass
class MutationResult:
    success: bool
    message: str
    data: Optional[Any]


@dataclass
class RoutingData:
    id: uuid.UUID
    endpoint: uuid.UUID
    session: Optional[uuid.UUID]
    status: RouteStatus
    traffic_ratio: float
    created_at: datetime
    error_data: dict[str, Any]

    @classmethod
    def from_row(cls, row: RoutingRow) -> Self:
        return cls(
            id=row.id,
            endpoint=row.endpoint,
            session=row.session,
            status=row.status,
            traffic_ratio=row.traffic_ratio,
            created_at=row.created_at,
            error_data=row.error_data,
        )


@dataclass
class RuntimeVariantData:
    name: str
    human_readable_name: str


@dataclass
class EndpointData:
    id: uuid.UUID
    image: Optional[ImageData]
    domain: str
    project: uuid.UUID
    resource_group: str
    resource_slots: ResourceSlot
    url: str

    model: uuid.UUID
    model_definition_path: str | None
    model_mount_destination: Optional[str]

    created_user_id: uuid.UUID
    created_user_email: Optional[str]

    session_owner_id: uuid.UUID
    session_owner_email: str

    tag: Optional[str]

    startup_command: Optional[str]
    bootstrap_script: Optional[str]
    callback_url: Optional[yarl.URL]
    environ: Optional[Mapping[str, Any]]

    name: str

    resource_opts: Optional[Mapping[str, Any]]
    replicas: int

    cluster_mode: ClusterMode
    cluster_size: int
    open_to_public: bool
    created_at: datetime
    destroyed_at: datetime
    retries: int

    routings: Optional[list[RoutingData]]
    lifecycle_stage: str
    runtime_variant: RuntimeVariant

    @classmethod
    def from_row(cls, row: Optional[EndpointRow]) -> Optional[Self]:
        if row is None:
            return None
        return cls(
            id=row.id,
            image=row.image_row.to_dataclass(),
            domain=row.domain,
            project=row.project,
            resource_group=row.resource_group,
            resource_slots=row.resource_slots,
            url=row.url,
            model=row.model,
            model_definition_path=row.model_definition_path,
            model_mount_destination=row.model_mount_destination,
            created_user_id=row.created_user,
            created_user_email=row.created_user_row.email
            if row.created_user_row is not None
            else None,
            session_owner_id=row.session_owner,
            session_owner_email=row.session_owner_row.email,
            tag=row.tag,
            startup_command=row.startup_command,
            bootstrap_script=row.bootstrap_script,
            callback_url=row.callback_url,
            environ=row.environ,
            name=row.name,
            resource_opts=row.resource_opts,
            replicas=row.replicas,
            cluster_mode=ClusterMode(row.cluster_mode),
            cluster_size=row.cluster_size,
            open_to_public=row.open_to_public,
            created_at=row.created_at,
            destroyed_at=row.destroyed_at,
            retries=row.retries,
            routings=[RoutingData.from_row(routing) for routing in row.routings]
            if row.routings
            else None,
            lifecycle_stage=row.lifecycle_stage.name,
            runtime_variant=row.runtime_variant,
        )


@dataclass
class EndpointTokenData:
    id: uuid.UUID
    token: str
    endpoint: uuid.UUID
    session_owner: uuid.UUID
    domain: str
    project: uuid.UUID
    created_at: datetime


@dataclass
class EndpointAutoScalingRuleData:
    id: uuid.UUID
    metric_source: AutoScalingMetricSource
    metric_name: str
    threshold: str
    comparator: AutoScalingMetricComparator
    step_size: int
    cooldown_seconds: int
    min_replicas: int
    max_replicas: int
    created_at: datetime
    last_triggered_at: datetime
    endpoint: uuid.UUID

    @classmethod
    def from_row(cls, row: Optional[EndpointAutoScalingRuleRow]) -> Optional[Self]:
        if row is None:
            return None
        return cls(
            id=row.id,
            metric_source=row.metric_source,
            metric_name=row.metric_name,
            threshold=row.threshold,
            comparator=row.comparator,
            step_size=row.step_size,
            cooldown_seconds=row.cooldown_seconds,
            min_replicas=row.min_replicas,
            max_replicas=row.max_replicas,
            created_at=row.created_at,
            last_triggered_at=row.last_triggered_at,
            endpoint=row.endpoint,
        )


@dataclass
class ModelServiceCreator(Creator):
    service_name: str
    replicas: int
    image: str
    runtime_variant: RuntimeVariant
    architecture: str
    group_name: str
    domain_name: str
    cluster_size: int
    cluster_mode: ClusterMode
    open_to_public: bool
    config: ServiceConfig
    sudo_session_enabled: bool
    model_service_prepare_ctx: ModelServicePrepareCtx
    tag: Optional[str] = None
    startup_command: Optional[str] = None
    bootstrap_script: Optional[str] = None
    callback_url: Optional[AnyUrl] = None

    @override
    def fields_to_store(self) -> dict[str, Any]:
        return {}


@dataclass
class ImageRef:
    name: str
    registry: OptionalState[str] = field(default_factory=OptionalState[str].nop)
    architecture: OptionalState[str] = field(default_factory=OptionalState[str].nop)


@dataclass
class ExtraMount:
    vfolder_id: OptionalState[uuid.UUID] = field(default_factory=OptionalState[uuid.UUID].nop)
    mount_destination: OptionalState[str] = field(default_factory=OptionalState[str].nop)
    type: OptionalState[str] = field(default_factory=OptionalState[str].nop)
    permission: OptionalState[str] = field(default_factory=OptionalState[str].nop)


@dataclass
class EndpointModifier(PartialModifier):
    resource_slots: OptionalState[ResourceSlot] = field(
        default_factory=OptionalState[ResourceSlot].nop
    )
    resource_opts: TriState[dict[str, Any]] = field(default_factory=TriState[dict[str, Any]].nop)
    cluster_mode: OptionalState[ClusterMode] = field(default_factory=OptionalState[ClusterMode].nop)
    cluster_size: OptionalState[int] = field(default_factory=OptionalState[int].nop)
    replicas: OptionalState[int] = field(default_factory=OptionalState[int].nop)
    desired_session_count: OptionalState[int] = field(default_factory=OptionalState.nop)
    image: TriState[ImageRef] = field(default_factory=TriState.nop)
    name: OptionalState[str] = field(default_factory=OptionalState.nop)
    resource_group: OptionalState[str] = field(default_factory=OptionalState[str].nop)
    model_definition_path: TriState[str] = field(default_factory=TriState[str].nop)
    open_to_public: OptionalState[bool] = field(default_factory=OptionalState[bool].nop)
    extra_mounts: OptionalState[list[ExtraMount]] = field(
        default_factory=OptionalState[list[ExtraMount]].nop
    )
    environ: TriState[dict[str, str]] = field(default_factory=TriState[dict[str, str]].nop)
    runtime_variant: OptionalState[RuntimeVariant] = field(
        default_factory=OptionalState[RuntimeVariant].nop
    )

    @override
    def fields_to_update(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.resource_slots.update_dict(to_update, "resource_slots")
        self.resource_opts.update_dict(to_update, "resource_opts")
        self.cluster_mode.update_dict(to_update, "cluster_mode")
        self.cluster_size.update_dict(to_update, "cluster_size")
        self.model_definition_path.update_dict(to_update, "model_definition_path")
        self.runtime_variant.update_dict(to_update, "runtime_variant")
        self.resource_group.update_dict(to_update, "resource_group")
        return to_update

    def fields_to_update_require_none_check(self) -> dict[str, Any]:
        # This method is used to update fields that require a check for None values
        to_update: dict[str, Any] = {}
        # The order of replicas and desired_session_count is important
        # as desired_session_count is legacy field and value of replicas need to override it
        self.desired_session_count.update_dict(to_update, "desired_session_count")
        self.replicas.update_dict(to_update, "replicas")
        self.environ.update_dict(to_update, "environ")
        return to_update


@dataclass
class EndpointAutoScalingRuleCreator(Creator):
    metric_source: AutoScalingMetricSource
    metric_name: str
    threshold: str
    comparator: AutoScalingMetricComparator
    step_size: int
    cooldown_seconds: int
    min_replicas: Optional[int] = None
    max_replicas: Optional[int] = None

    @override
    def fields_to_store(self) -> dict[str, Any]:
        return {
            "metric_source": self.metric_source,
            "metric_name": self.metric_name,
            "threshold": self.threshold,
            "comparator": self.comparator,
            "step_size": self.step_size,
            "cooldown_seconds": self.cooldown_seconds,
            "min_replicas": self.min_replicas,
            "max_replicas": self.max_replicas,
        }


@dataclass
class EndpointAutoScalingRuleModifier(PartialModifier):
    metric_source: OptionalState[AutoScalingMetricSource] = field(default_factory=OptionalState.nop)
    metric_name: OptionalState[str] = field(default_factory=OptionalState.nop)
    threshold: OptionalState[Decimal] = field(default_factory=OptionalState.nop)
    comparator: OptionalState[AutoScalingMetricComparator] = field(
        default_factory=OptionalState.nop
    )
    step_size: OptionalState[int] = field(default_factory=OptionalState.nop)
    cooldown_seconds: OptionalState[int] = field(default_factory=OptionalState.nop)
    min_replicas: TriState[int] = field(default_factory=TriState.nop)
    max_replicas: TriState[int] = field(default_factory=TriState.nop)

    @override
    def fields_to_update(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.metric_source.update_dict(to_update, "metric_source")
        self.metric_name.update_dict(to_update, "metric_name")
        self.threshold.update_dict(to_update, "threshold")
        self.comparator.update_dict(to_update, "comparator")
        self.step_size.update_dict(to_update, "step_size")
        self.cooldown_seconds.update_dict(to_update, "cooldown_seconds")
        self.min_replicas.update_dict(to_update, "min_replicas")
        self.max_replicas.update_dict(to_update, "max_replicas")
        return to_update


class ImageEnvironment(BaseModel):
    image: str = Field(
        description="""
        Container image to use for the model service.
        """,
        examples=[
            "myregistry/myimage:latest",
        ],
    )
    architecture: str = Field(
        description="""
        Architecture of the container image.
        """,
        examples=[
            "x86_64",
            "arm64",
        ],
    )


class ModelServiceDefinition(BaseModel):
    environment: Optional[ImageEnvironment] = Field(
        default=None,
        description="""
        Environment in which the model service will run.
        """,
        examples=[
            {
                "image": "myregistry/myimage:latest",
                "architecture": "x86_64",
            }
        ],
    )
    resource_slots: Optional[dict[str, Any]] = Field(
        default=None,
        description="""
        Resource slots used by the model service session.
        """,
        examples=[
            {"cpu": 1, "mem": "2gb"},
        ],
    )
    environ: Optional[dict[str, str]] = Field(
        default=None,
        description="""
        Environment variables to set for the model service.
        """,
        examples=[
            {"MY_ENV_VAR": "value", "ANOTHER_VAR": "another_value"},
        ],
    )



import uuid
from dataclasses import dataclass
from typing import Optional, override

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.image.types import ImageData
from ai.backend.manager.models.user import UserRole
from ai.backend.manager.services.image.actions.base import ImageAction


@dataclass
class ForgetImageAction(ImageAction):
    user_id: uuid.UUID
    client_role: UserRole
    reference: str
    architecture: str

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "forget"


@dataclass
class ForgetImageActionResult(BaseActionResult):
    image: ImageData

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.image.id)


class ForgetImageActionGenericForbiddenError(BackendAIError):
    error_type = "https://api.backend.ai/probs/generic-forbidden"
    error_title = "Access to this resource is forbidden."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.IMAGE,
            operation=ErrorOperation.SOFT_DELETE,
            error_detail=ErrorDetail.FORBIDDEN,
        )



import asyncio
import decimal
import logging
from typing import Any, Awaitable, Callable

import sqlalchemy as sa
from sqlalchemy.orm.exc import NoResultFound

from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.models.endpoint import (
    EndpointAutoScalingRuleRow,
    EndpointLifecycle,
    EndpointRow,
)
from ai.backend.manager.models.user import UserRole
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine, execute_with_retry
from ai.backend.manager.services.model_serving.actions.create_auto_scaling_rule import (
    CreateEndpointAutoScalingRuleAction,
    CreateEndpointAutoScalingRuleActionResult,
)
from ai.backend.manager.services.model_serving.actions.delete_auto_scaling_rule import (
    DeleteEndpointAutoScalingRuleAction,
    DeleteEndpointAutoScalingRuleActionResult,
)
from ai.backend.manager.services.model_serving.actions.modify_auto_scaling_rule import (
    ModifyEndpointAutoScalingRuleAction,
    ModifyEndpointAutoScalingRuleActionResult,
)
from ai.backend.manager.services.model_serving.actions.scale_service_replicas import (
    ScaleServiceReplicasAction,
    ScaleServiceReplicasActionResult,
)
from ai.backend.manager.services.model_serving.exceptions import (
    EndpointAutoScalingRuleNotFound,
    EndpointNotFound,
    GenericForbidden,
    InvalidAPIParameters,
    ModelServiceNotFound,
)
from ai.backend.manager.services.model_serving.services.utils import verify_user_access_scopes
from ai.backend.manager.services.model_serving.types import (
    EndpointAutoScalingRuleData,
    MutationResult,
)

log = BraceStyleAdapter(logging.getLogger(__name__))


class AutoScalingService:
    _db: ExtendedAsyncSAEngine

    def __init__(
        self,
        db: ExtendedAsyncSAEngine,
    ) -> None:
        self._db = db

    async def scale_service_replicas(
        self, action: ScaleServiceReplicasAction
    ) -> ScaleServiceReplicasActionResult:
        async with self._db.begin_readonly_session() as db_sess:
            try:
                endpoint = await EndpointRow.get(db_sess, action.service_id, load_routes=True)
            except NoResultFound:
                raise ModelServiceNotFound
        await verify_user_access_scopes(self._db, action.requester_ctx, endpoint.session_owner)

        async with self._db.begin_session() as db_sess:
            query = (
                sa.update(EndpointRow)
                .where(EndpointRow.id == action.service_id)
                .values({"replicas": action.to})
            )
            await db_sess.execute(query)
            return ScaleServiceReplicasActionResult(
                current_route_count=len(endpoint.routings), target_count=action.to
            )

    async def create_endpoint_auto_scaling_rule(
        self, action: CreateEndpointAutoScalingRuleAction
    ) -> CreateEndpointAutoScalingRuleActionResult:
        async with self._db.begin_session(commit_on_end=True) as db_session:
            try:
                row = await EndpointRow.get(db_session, action.endpoint_id)
                if row.lifecycle_stage in EndpointLifecycle.inactive_states():
                    raise EndpointNotFound
            except NoResultFound:
                raise EndpointNotFound

            match action.requester_ctx.user_role:
                case UserRole.SUPERADMIN:
                    pass
                case UserRole.ADMIN:
                    if row.domain != action.requester_ctx.domain_name:
                        raise GenericForbidden
                case UserRole.USER:
                    if row.created_user != action.requester_ctx.user_id:
                        raise GenericForbidden

            try:
                _threshold = decimal.Decimal(action.creator.threshold)
            except decimal.InvalidOperation:
                raise InvalidAPIParameters(f"Cannot convert {action.creator.threshold} to Decimal")

            async def _do_mutate() -> MutationResult:
                created_rule = await row.create_auto_scaling_rule(
                    db_session,
                    action.creator.metric_source,
                    action.creator.metric_name,
                    _threshold,
                    action.creator.comparator,
                    action.creator.step_size,
                    cooldown_seconds=action.creator.cooldown_seconds,
                    min_replicas=action.creator.min_replicas,
                    max_replicas=action.creator.max_replicas,
                )
                return MutationResult(
                    success=True,
                    message="Auto scaling rule created",
                    data=created_rule,
                )

            res = await self._db_mutation_wrapper(_do_mutate)

            return CreateEndpointAutoScalingRuleActionResult(
                success=res.success,
                data=EndpointAutoScalingRuleData.from_row(res.data),
            )

    async def modify_endpoint_auto_scaling_rule(
        self, action: ModifyEndpointAutoScalingRuleAction
    ) -> ModifyEndpointAutoScalingRuleActionResult:
        async with self._db.begin_session(commit_on_end=True) as db_session:
            try:
                row = await EndpointAutoScalingRuleRow.get(
                    db_session, action.id, load_endpoint=True
                )
                if row.endpoint_row.lifecycle_stage in EndpointLifecycle.inactive_states():
                    raise EndpointAutoScalingRuleNotFound
            except NoResultFound:
                raise EndpointAutoScalingRuleNotFound

            match action.requester_ctx.user_role:
                case UserRole.SUPERADMIN:
                    pass
                case UserRole.ADMIN:
                    if row.endpoint_row.domain != action.requester_ctx.domain_name:
                        raise GenericForbidden
                case UserRole.USER:
                    if row.endpoint_row.created_user != action.requester_ctx.user_id:
                        raise GenericForbidden

            async def _do_mutate() -> MutationResult:
                fields_to_update: dict[str, Any] = action.modifier.fields_to_update()
                for key, value in fields_to_update.items():
                    setattr(row, key, value)

                return MutationResult(
                    success=True,
                    message="Auto scaling rule updated",
                    data=row,
                )

            res = await self._db_mutation_wrapper(_do_mutate)

            return ModifyEndpointAutoScalingRuleActionResult(
                success=res.success,
                data=EndpointAutoScalingRuleData.from_row(res.data),
            )

    async def delete_endpoint_auto_scaling_rule(
        self, action: DeleteEndpointAutoScalingRuleAction
    ) -> DeleteEndpointAutoScalingRuleActionResult:
        async with self._db.begin_session(commit_on_end=True) as db_session:
            try:
                row = await EndpointAutoScalingRuleRow.get(
                    db_session, action.id, load_endpoint=True
                )
            except NoResultFound:
                raise EndpointAutoScalingRuleNotFound

            match action.requester_ctx.user_role:
                case UserRole.SUPERADMIN:
                    pass
                case UserRole.ADMIN:
                    if row.endpoint_row.domain != action.requester_ctx.domain_name:
                        raise GenericForbidden
                case UserRole.USER:
                    if row.endpoint_row.created_user != action.requester_ctx.user_id:
                        raise GenericForbidden

            async def _do_mutate() -> MutationResult:
                await db_session.delete(row)
                return MutationResult(
                    success=True,
                    message="Auto scaling rule removed",
                    data=None,
                )

            res = await self._db_mutation_wrapper(_do_mutate)

            return DeleteEndpointAutoScalingRuleActionResult(
                success=res.success,
            )

    async def _db_mutation_wrapper(
        self, _do_mutate: Callable[[], Awaitable[MutationResult]]
    ) -> MutationResult:
        try:
            return await execute_with_retry(_do_mutate)
        except sa.exc.IntegrityError as e:
            log.warning("db_mutation_wrapper(): integrity error ({})", repr(e))
            return MutationResult(success=False, message=f"integrity error: {e}", data=None)
        except sa.exc.StatementError as e:
            log.warning(
                "db_mutation_wrapper(): statement error ({})\n{}",
                repr(e),
                e.statement or "(unknown)",
            )
            orig_exc = e.orig
            return MutationResult(success=False, message=str(orig_exc), data=None)
        except (asyncio.CancelledError, asyncio.TimeoutError):
            raise
        except Exception:
            log.exception("db_mutation_wrapper(): other error")
            raise



from __future__ import annotations

import asyncio
import contextlib
import os
import sys
from pathlib import Path
from pprint import pformat
from typing import AsyncIterator, Optional, Self

import attrs
import click

from ai.backend.common import redis_helper
from ai.backend.common.config import find_config_file
from ai.backend.common.defs import (
    REDIS_IMAGE_DB,
    REDIS_LIVE_DB,
    REDIS_STATISTICS_DB,
    REDIS_STREAM_DB,
    RedisRole,
)
from ai.backend.common.etcd import AsyncEtcd, ConfigScopes
from ai.backend.common.exception import ConfigurationError
from ai.backend.common.types import RedisConnectionInfo, RedisProfileTarget
from ai.backend.logging import AbstractLogger, LocalLogger, LogLevel
from ai.backend.manager.config.bootstrap import BootstrapConfig
from ai.backend.manager.config.loader.legacy_etcd_loader import LegacyEtcdLoader
from ai.backend.manager.config.unified import ManagerUnifiedConfig, RedisConfig


class CLIContext:
    _bootstrap_config: Optional[BootstrapConfig]
    _logger: AbstractLogger

    def __init__(self, log_level: LogLevel, config_path: Optional[Path] = None) -> None:
        self.config_path = config_path
        self.log_level = log_level
        self._bootstrap_config = None

    def get_bootstrap_config(self) -> BootstrapConfig:
        # Lazy-load the configuration only when requested.
        try:
            if self._bootstrap_config is None:
                if self.config_path is None:
                    self.config_path = find_config_file("manager")

                self._bootstrap_config = asyncio.run(
                    BootstrapConfig.load_from_file(self.config_path, self.log_level)
                )
        except ConfigurationError as e:
            print(
                "ConfigurationError: Could not read or validate the manager local config:",
                file=sys.stderr,
            )
            print(pformat(e.invalid_data), file=sys.stderr)
            raise click.Abort()
        return self._bootstrap_config

    def __enter__(self) -> Self:
        # The "start-server" command is injected by ai.backend.cli from the entrypoint
        # and it has its own multi-process-aware logging initialization.
        # If we duplicate the local logging with it, the process termination may hang.
        click_ctx = click.get_current_context()
        if click_ctx.invoked_subcommand != "start-server":
            logging_config = {
                "level": self.log_level,
                "pkg-ns": {
                    "": LogLevel.WARNING,
                    "ai.backend": self.log_level,
                },
            }
            try:
                # Try getting the logging config but silently fallback to the default if not
                # present (e.g., when `mgr gql show` command used in CI without installation as
                # addressed in #1686).
                with open(os.devnull, "w") as sink, contextlib.redirect_stderr(sink):
                    logging_config = self.get_bootstrap_config().logging
            except click.Abort:
                pass
            self._logger = LocalLogger(logging_config)
            self._logger.__enter__()
        return self

    def __exit__(self, *exc_info) -> None:
        click_ctx = click.get_current_context()
        if click_ctx.invoked_subcommand != "start-server":
            self._logger.__exit__()


@contextlib.asynccontextmanager
async def etcd_ctx(cli_ctx: CLIContext) -> AsyncIterator[AsyncEtcd]:
    etcd_config = cli_ctx.get_bootstrap_config().etcd
    creds = None
    if etcd_config.user:
        if not etcd_config.password:
            raise ConfigurationError({
                "etcd": "password is required when user is set",
            })

        creds = {
            "user": etcd_config.user,
            "password": etcd_config.password,
        }
    scope_prefix_map = {
        ConfigScopes.GLOBAL: "",
        # TODO: provide a way to specify other scope prefixes
    }
    etcd = AsyncEtcd(
        etcd_config.addr.to_legacy(),
        etcd_config.namespace,
        scope_prefix_map,
        credentials=creds,
    )
    try:
        yield etcd
    finally:
        await etcd.close()


@contextlib.asynccontextmanager
async def config_ctx(cli_ctx: CLIContext) -> AsyncIterator[ManagerUnifiedConfig]:
    # scope_prefix_map is created inside ConfigServer

    etcd = AsyncEtcd.initialize(cli_ctx.get_bootstrap_config().etcd.to_dataclass())
    etcd_loader = LegacyEtcdLoader(etcd)
    redis_config = await etcd_loader.load()
    unified_config = ManagerUnifiedConfig(**redis_config)

    try:
        yield unified_config
    finally:
        await etcd_loader.close()


@attrs.define(auto_attribs=True, frozen=True, slots=True)
class RedisConnectionSet:
    live: RedisConnectionInfo
    stat: RedisConnectionInfo
    image: RedisConnectionInfo
    stream: RedisConnectionInfo


@contextlib.asynccontextmanager
async def redis_ctx(cli_ctx: CLIContext) -> AsyncIterator[RedisConnectionSet]:
    etcd = AsyncEtcd.initialize(cli_ctx.get_bootstrap_config().etcd.to_dataclass())
    loader = LegacyEtcdLoader(etcd, config_prefix="config/redis")
    raw_redis_config = await loader.load()
    redis_config = RedisConfig(**raw_redis_config)
    etcd_redis_config = RedisProfileTarget.from_dict(redis_config.model_dump())

    redis_live = redis_helper.get_redis_object(
        etcd_redis_config.profile_target(RedisRole.LIVE),
        name="mgr_cli.live",
        db=REDIS_LIVE_DB,
    )
    redis_stat = redis_helper.get_redis_object(
        etcd_redis_config.profile_target(RedisRole.STATISTICS),
        name="mgr_cli.stat",
        db=REDIS_STATISTICS_DB,
    )
    redis_image = redis_helper.get_redis_object(
        etcd_redis_config.profile_target(RedisRole.IMAGE),
        name="mgr_cli.image",
        db=REDIS_IMAGE_DB,
    )
    redis_stream = redis_helper.get_redis_object(
        etcd_redis_config.profile_target(RedisRole.STREAM),
        name="mgr_cli.stream",
        db=REDIS_STREAM_DB,
    )
    yield RedisConnectionSet(
        live=redis_live,
        stat=redis_stat,
        image=redis_image,
        stream=redis_stream,
    )
    await redis_stream.close()
    await redis_image.close()
    await redis_stat.close()
    await redis_live.close()
